package be.iminds.ilabt.jfed.ui.cli2;

import be.iminds.ilabt.jfed.ui.cli2.instruction.Instruction;
import be.iminds.ilabt.jfed.ui.cli2.instruction.InstructionTypeIdResolver;
import be.iminds.ilabt.jfed.ui.cli2.wizard.ActionWizard;
import be.iminds.ilabt.jfed.ui.commandline.BaseCli;
import be.iminds.ilabt.jfed.util.common.DefaultOmittingSerializationUtil;
import be.iminds.ilabt.jfed.util.common.IOUtils;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.fasterxml.jackson.databind.SequenceWriter;
import com.fasterxml.jackson.databind.SerializationFeature;
import com.fasterxml.jackson.dataformat.yaml.YAMLFactory;
import com.fasterxml.jackson.dataformat.yaml.YAMLGenerator;
import javafx.embed.swing.JFXPanel;
import org.apache.commons.cli.HelpFormatter;
import org.apache.commons.cli.Options;
import org.slf4j.LoggerFactory;

import java.io.*;
import java.lang.reflect.Constructor;
import java.util.Map;

/**
 * ExperimenterCli
 */
@SuppressWarnings("AccessStaticViaInstance")
public class ExperimenterCli {
    private final static org.slf4j.Logger LOG = LoggerFactory.getLogger(ExperimenterCli.class);

    private static final String KEY_EXPIRATION_TIME = "expiration-time";
    private static final String KEY_REQUEST_RSPEC = "request-rspec";

    public static void main(String[] args) throws IOException, ClassNotFoundException {
        int retval = main_helper(args, System.out, System.err, System.in, new BaseCli(System.err), true);
        System.exit(retval);
    }

    public static int main_helper(String[] args,
                                  final PrintStream out, final PrintStream err, final InputStream in,
                                  BaseCli baseCli,
                                  boolean temporarilyRedirectSystemStreams) throws IOException, ClassNotFoundException {
        final PrintStream orig_out = System.out;
        final PrintStream orig_err = System.err;
        final InputStream orig_in = System.in;
        if (temporarilyRedirectSystemStreams) {
            System.setOut(out);
            System.setErr(err);
            System.setIn(in);
        }

        int retval = -1;
        try {
            retval = main_internal(args, out, err, in, baseCli);
        } catch (CliExitException e) {
            System.err.println(e.getExitReason());
            return e.getExitCode();
        } finally {
            if (temporarilyRedirectSystemStreams) {
                System.setOut(orig_out);
                System.setErr(orig_err);
                System.setIn(orig_in);
            }
        }
        return retval;
    }

    public static int main_internal(String[] args,
                                    final PrintStream out, final PrintStream err, final InputStream in,
                                    BaseCli baseCli) throws CliExitException {
        //initialize JavaFX-toolkit
        new JFXPanel();

        final ExperimenterCommandLineParser experimenterCommandLineParser;

        experimenterCommandLineParser = new ExperimenterCommandLineParser(args, out, err, in, baseCli);
        ExperimenterCommandLineParser.GenericCliArguments genCliArgs = experimenterCommandLineParser.getGenericCliArguments();

        if (args.length == 0) {
            err.println("Invalid syntax: No argument.\n" +
                    "Try:\n" +
                    "   --help     ->  generic help\n" +
                    "   --wizard   ->  generate an instruction  (recommended to get started)");
        }
        if (genCliArgs.isWizard()) {
            ActionWizard wizard = new ActionWizard(out, err, in);
            try {
                Instruction instr = wizard.generate();

                Map generatedInstructionMap;
                ObjectMapper mapper;
                try {
                    ObjectMapper om = new ObjectMapper();
                    generatedInstructionMap = om.convertValue(instr, Map.class);

                    out.println("\n*** Action file generation DONE ***\n");
                    boolean minimizeInstruction = wizard.askBool(
                            "Minimize the generated yml by skipping defaults", true);
                    if (minimizeInstruction) {
                        Instruction defaultInstruction = getDefaultInstruction(instr.getAction());
                        Map defaultInstructionMap = om.convertValue(defaultInstruction, Map.class);
                        defaultInstructionMap.remove("action"); //action should never be seen as a default
                        generatedInstructionMap = DefaultOmittingSerializationUtil.omitDefault(
                                generatedInstructionMap,
                                defaultInstructionMap
                        );
                    }

                    YAMLFactory yf = new YAMLFactory();
                    yf.enable(YAMLGenerator.Feature.MINIMIZE_QUOTES);
                    mapper = new ObjectMapper(yf);
                } catch (Exception e) {
                    throw new CliExitException(Context.ExitReason.ERROR, "Error while converting generated config to yml: "+e.getMessage(), e);
                }

                out.println("What to do with the generated action file?");
                boolean saveToFile = wizard.askBoolChoice("Save to a file", "Print", true);
                String filename;
                PrintStream outStream;
                if (saveToFile) {
                    filename = wizard.askString("Filename to save to", "action.yml", false);
                    outStream = new PrintStream(new FileOutputStream(filename));
                } else {
                    filename = "<my_generated_action.yml>";
                    outStream = out;
                }
                try (SequenceWriter sw = mapper.writerWithDefaultPrettyPrinter()
                        .withFeatures(SerializationFeature.ORDER_MAP_ENTRIES_BY_KEYS)
                        .writeValues(outStream)) {
                    if (!saveToFile) {
                        out.println("\nGenerated Action yml:");
                    }
                    sw.write(generatedInstructionMap);

                    if (!saveToFile) {
                        out.println("\n\nWrite it to a file and execute it with:\n");
                    } else {
                        out.println("\n\nYou can run the saved action with:\n");
                    }

                    out.println("    java -jar experimenter-cli2.jar -a "+filename);
                } catch (IOException e) {
                    throw new CliExitException(Context.ExitReason.ERROR, "Error while converting generated config to yml: "+e.getMessage(), e);
                }
                return 0;
            } catch (IOException e) {
                err.println("Exception while running wizard:");
                e.printStackTrace();
                return -1;
            }
        }

        if (/*args.length == 0 || */(genCliArgs.isHelp() && !genCliArgs.hasAction())) {
            experimenterCommandLineParser.generalHelp(err);
            return 0;
        }
        if (genCliArgs.isVersion()) {
            experimenterCommandLineParser.showVersion(out);
            return 0;
        }

        if (genCliArgs.isDebug())
            Context.setupDebugOutputToStdout();

        ExperimenterCommandLineParser.ProcessInstructionResult processInstructionResult =
                new ExperimenterCommandLineParser.ProcessInstructionResult();
        try {
            experimenterCommandLineParser.processInstruction(processInstructionResult);
//        } catch (CliExitException e) {
        } catch (Exception e) {
            if (!genCliArgs.isHelp() || processInstructionResult.action == null) {
                //outputs not yet OK, so show error on stdout
                err.println("" + e.getMessage());
                throw e;
            } else {
                //if help, just continue
            }
        }
        Instruction.Action action = processInstructionResult.action;
        Instruction instruction = processInstructionResult.instruction;
        Options options = processInstructionResult.options;
        assert action != null;

        if (genCliArgs.isHelp()) {
            if (options != null) {
                HelpFormatter formatter = new HelpFormatter();
                formatter.setWidth(IOUtils.getTermWidth(80));
                formatter.printHelp("experimenter-cli --action " + action.getNiceName(), options);
            } else {
                err.println("Error getting option info for \"experimenter-cli " +
                        "--action " + action.getNiceName()+"\"");
            }
            out.flush();
            err.flush();

            Map demoInstructionMap;
            ObjectMapper mapper;
            try {
                Class<? extends Instruction> instructionClass = InstructionTypeIdResolver.classFromAction(action);
                assert instructionClass != null;
                ObjectMapper om = new ObjectMapper();
                Instruction demoInstruction = getDemoInstruction(action);
                demoInstructionMap = om.convertValue(demoInstruction, Map.class);

                YAMLFactory yf = new YAMLFactory();
                yf.enable(YAMLGenerator.Feature.MINIMIZE_QUOTES);
                mapper = new ObjectMapper(yf);
            } catch (CliExitException e) {
                throw e;
            } catch (Exception e) {
                throw new CliExitException(Context.ExitReason.ERROR, "Error while converting default config to yml: "+e.getMessage(), e);
            }
            try (SequenceWriter sw = mapper.writerWithDefaultPrettyPrinter()
                    .withFeatures(SerializationFeature.ORDER_MAP_ENTRIES_BY_KEYS)
                    .writeValues(out)) {
                out.println("\nExample yml for \""+action.getNiceName()+"\":");
                sw.write(demoInstructionMap);
            } catch (IOException e) {
                throw new CliExitException(Context.ExitReason.ERROR, "Error while converting default config to yml: "+e.getMessage(), e);
            }
            return 0;
        }

        if (genCliArgs.isPrintAction()) {
            ObjectMapper om = new ObjectMapper();
            Instruction defaultInstruction = getDefaultInstruction(action);
            Map defaultInstructionMap = om.convertValue(defaultInstruction, Map.class);
            defaultInstructionMap.remove("action"); //action should never be seen as a default
            Map prunedInstructions = DefaultOmittingSerializationUtil.omitDefault(
                    om.convertValue(instruction, Map.class),
                    defaultInstructionMap
            );

            YAMLFactory yf = new YAMLFactory();
            yf.enable(YAMLGenerator.Feature.MINIMIZE_QUOTES);
            ObjectMapper mapper = new ObjectMapper(yf);
            try (SequenceWriter sw = mapper.writerWithDefaultPrettyPrinter()
                    .withFeatures(SerializationFeature.ORDER_MAP_ENTRIES_BY_KEYS)
                    .writeValues(out)) {
                sw.write(prunedInstructions);
            } catch (IOException e) {
                throw new CliExitException(Context.ExitReason.ERROR, "Error while converting config to yml: "+e.getMessage(), e);
            }
            return 0;
        }

        Context context = new Context(args, genCliArgs, instruction, out, err, in, baseCli);

        try {
            //setup the outputs according to the instruction
            context.setupDebugOutput();
            context.setupActionOutput();
        } catch (CliExitException e) {
            //outputs not yet OK, so show error on stdout
            err.println(""+e.getMessage());
            throw e;
        }

        try {
            //note: thanks to context.setupDebugOutput() we can use default logging with static loggers
            LOG.debug("Final Instruction: " + context.getInstruction());

            context.setupInjector();
            context.setupProxy();
            context.setupCallOutput(); //needs Logger  (so take injector into account here)

            InstructionExecutor instructionExecutor = new InstructionExecutor(context);
            instructionExecutor.run(context, instruction);
            return 0;
        } catch (CliExitException e) {
            assert e.getExitReason() != Context.ExitReason.SUCCESS;
            if (e.getExitReason() != Context.ExitReason.SUCCESS && e.getMessage() != null) {
                context.getActionLogger().errorPrintln(e.getMessage());
            }
            return e.getExitCode();
        } finally {
            context.closeOutputs();
        }
    }

    private static Instruction getDemoInstruction(Instruction.Action action) throws CliExitException {
        Class<? extends Instruction> instructionClass = InstructionTypeIdResolver.classFromAction(action);
        assert instructionClass != null;
        ObjectMapper om = new ObjectMapper();
        Instruction defaultInstruction = null;
        try {
            Constructor<? extends Instruction> constructor = instructionClass.getDeclaredConstructor(Boolean.TYPE);
            constructor.setAccessible(true);
            return constructor.newInstance(true);
        } catch (Exception e) {
            throw new CliExitException(Context.ExitReason.ERROR,
                    "Implementation exception for "+instructionClass.getName()+" (demo constructor missing?)", e);
        }
    }

    private static Instruction getDefaultInstruction(Instruction.Action action) throws CliExitException {
        Class<? extends Instruction> instructionClass = InstructionTypeIdResolver.classFromAction(action);
        assert instructionClass != null;
        ObjectMapper om = new ObjectMapper();
        Instruction defaultInstruction = null;
        try {
            Constructor<? extends Instruction> constructor = instructionClass.getDeclaredConstructor(Boolean.TYPE);
            constructor.setAccessible(true);
            return constructor.newInstance(false);
        } catch (Exception e) {
            throw new CliExitException(Context.ExitReason.ERROR,
                    "Implementation exception for "+instructionClass.getName()+" (default constructor missing?)", e);
        }
    }

//    private static void addToUserSpecs(List<UserSpec> userSpecs, GeniUrn userUrn, String sshKey) {
//        assert userUrn != null;
//        addToUserSpecs(userSpecs, userUrn, Collections.singletonList(sshKey));
//    }
//    private static void addToUserSpecs(List<UserSpec> userSpecs, String userUrnString, String sshKey) {
//        addToUserSpecs(userSpecs, userUrnString, Collections.singletonList(sshKey));
//    }
//    private static void addToUserSpecs(List<UserSpec> userSpecs, GeniUrn userUrn, Collection<String> sshKeys) {
//        assert userUrn != null;
//        addToUserSpecs(userSpecs, userUrn.toString(), sshKeys);
//    }
//    private static void addToUserSpecs(List<UserSpec> userSpecs, String userUrnString, Collection<String> sshKeys) {
//        if (sshKeys == null || sshKeys.isEmpty())
//            return;
//        for (UserSpec us : userSpecs) {
//            if (Objects.equals(us.getUrn(), userUrnString)) {
//                //TODO (low priority) make this work for SSH keys that differ only in comment, but not in actual key
//                Set<String> newSskKeys = new HashSet<>(sshKeys);
//                newSskKeys.removeAll(us.getSshKey());
//                if (!newSskKeys.isEmpty())
//                    us.getSshKey().addAll(newSskKeys);
//                return;
//            }
//        }
//
//        UserSpec newUs = new UserSpec(userUrnString, sshKeys);
//        userSpecs.add(newUs);
//    }
}
